#ifndef KERN26_IFACE
#include <linux/wrapper.h>
#endif
#include <linux/vmalloc.h>
#include <linux/videodev.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/module.h>

struct uscam_mdev_info {
  int xsize, ysize;
  int framelen;
  int format;
} mdev_info;

void *framebuf;
int has_frame;
wait_queue_head_t frameq;
int activated;

#ifdef KERN26_IFACE
int uscam_video_open(struct inode *inode, struct file *file) {
  struct video_device *vdev = video_devdata(file);
  file->private_data = vdev;
  return 0;
}

int uscam_video_close(struct inode *inode, struct file *file) {
  return 0;
}

ssize_t uscam_video_read(struct file *file, char *buf,
	size_t count, loff_t *ppos ) {
  return 0;
}

unsigned int uscam_video_poll(struct file *file, poll_table *wait) {
  return 0;
}
#else
int uscam_video_open(struct video_device *vdev, int mode) {
  return 0;
}

void uscam_video_close(struct video_device *vdev) {
}

long uscam_video_read(struct video_device *vdev, char *buf, unsigned long count, int noblock) {
  return 0;
}

unsigned int uscam_video_poll(struct video_device *vdev, struct file *file, poll_table *wait) {
  return 0;
}
#endif

#ifdef KERN26_IFACE
int uscam_video_do_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, void *arg );

int uscam_video_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg ) {
  return video_usercopy(inode, file, cmd, arg, uscam_video_do_ioctl);
}

int uscam_video_do_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, void *arg ) {
  struct video_device *vdev = file->private_data;
#else
int uscam_video_ioctl(struct video_device *vdev, unsigned int cmd, void *arg ) {
#endif
  DECLARE_WAITQUEUE(wait, current);

  switch (cmd) {
  case VIDIOCGCAP: {
#ifdef KERN26_IFACE
	struct video_capability *caps = arg;

	strcpy( caps->name, vdev->name );
	caps->type = VID_TYPE_CAPTURE;
	caps->channels = 1;
	caps->audios = 0;
	caps->minwidth = mdev_info.xsize;
	caps->minheight = mdev_info.ysize;
	caps->maxwidth = mdev_info.xsize;
	caps->maxheight = mdev_info.ysize;
	break;
#else
	struct video_capability caps;

	strcpy( caps.name, vdev->name );
	caps.type = VID_TYPE_CAPTURE;
	caps.channels = 1;
	caps.audios = 0;
	caps.minwidth = mdev_info.xsize;
	caps.minheight = mdev_info.ysize;
	caps.maxwidth = mdev_info.xsize;
	caps.maxheight = mdev_info.ysize;
	if ( copy_to_user(arg, &caps, sizeof(caps)) )
		return -EFAULT;
	break;
#endif
  }
  case 	VIDIOCGMBUF: {
#ifdef KERN26_IFACE
  	struct video_mbuf *vm = arg;

	memset( vm, 0, sizeof(struct video_mbuf) );
	vm->size = mdev_info.framelen;
	vm->frames = 1;
	vm->offsets[0] = 0;
	break;
#else
  	struct video_mbuf vm;

	memset( &vm, 0, sizeof(vm) );
	vm.size = mdev_info.framelen;
	vm.frames = 1;
	vm.offsets[0] = 0;
	if ( copy_to_user(arg, &vm, sizeof(vm)) )
		return -EFAULT;
	break;
#endif
  }
  case VIDIOCGWIN: {
#ifdef KERN26_IFACE
	struct video_window *vw = arg;

	memset( vw, 0, sizeof(struct video_window) );
	vw->x = 0;
	vw->y = 0;
	vw->width = mdev_info.xsize;
	vw->height = mdev_info.ysize;
	vw->chromakey = 0;
	vw->flags = 0;
	break;
#else
	struct video_window vw;

	memset( &vw, 0, sizeof(vw) );
	vw.x = 0;
	vw.y = 0;
	vw.width = mdev_info.xsize;
	vw.height = mdev_info.ysize;
	vw.chromakey = 0;
	vw.flags = 0;
	if ( copy_to_user(arg, &vw, sizeof(vw) ) )
		return -EFAULT;
	break;
#endif
  }
  case VIDIOCSWIN: {
#ifdef KERN26_IFACE
	struct video_window *vw = arg;

	if ( vw->width != mdev_info.xsize || vw->height != mdev_info.ysize )
		return -EINVAL;
	break;
#else
	struct video_window vw;

	if ( copy_from_user( &vw, arg, sizeof(vw) ) )
		return -EFAULT;
	if ( vw.width != mdev_info.xsize || vw.height != mdev_info.ysize )
		return -EINVAL;
	break;
#endif
  }
  case VIDIOCGPICT: {
#ifdef KERN26_IFACE
	struct video_picture *vp = arg;

	memset( vp, 0, sizeof(struct video_picture) );
	vp->palette = mdev_info.format;
	break;
#else
	struct video_picture vp;

	memset( &vp, 0, sizeof(vp) );
	vp.palette = mdev_info.format;
	if ( copy_to_user( arg, &vp, sizeof(vp) ) )
		return -EFAULT;
	break;
#endif
  }
  case VIDIOCSPICT:
	break;
  case VIDIOCMCAPTURE:
  	break;
  case VIDIOCSYNC: {
#ifdef KERN26_IFACE
  	int *frame = arg;

	if ( *frame != 0 ) return -EINVAL;
	add_wait_queue( &frameq, &wait );
	while ( !has_frame && !signal_pending(current) ) {
		set_current_state( TASK_INTERRUPTIBLE );
		schedule();
	}
	remove_wait_queue( &frameq, &wait );
	set_current_state( TASK_RUNNING );
	if ( signal_pending(current) ) return -ERESTARTSYS;

	has_frame = 0;
	break;
#else
  	int frame;

	if ( copy_from_user( &frame, arg, sizeof(frame) ) )
		return -EFAULT;
	if ( frame != 0 ) return -EINVAL;
	add_wait_queue( &frameq, &wait );
	while ( !has_frame && !signal_pending(current) ) {
		set_current_state( TASK_INTERRUPTIBLE );
		schedule();
	}
	remove_wait_queue( &frameq, &wait );
	set_current_state( TASK_RUNNING );
	if ( signal_pending(current) ) return -ERESTARTSYS;

	has_frame = 0;
	break;
#endif
  }
  default:
  	return -ENOIOCTLCMD;
  }

  return 0;
}

static inline unsigned long kvirt_to_pa(unsigned long adr) {
  unsigned long kva, ret;
   
  kva = (unsigned long) page_address(vmalloc_to_page((void *)adr));
  kva |= adr & (PAGE_SIZE-1); /* restore the offset */
  ret = __pa(kva);
  return ret;
}

static void * rvmalloc(unsigned long size) {
  void *mem;
  unsigned long adr;

  size=PAGE_ALIGN(size);
  mem=vmalloc_32(size);
  if (mem) {
	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr=(unsigned long) mem;
	while (size > 0) {
#ifdef KERN26_IFACE
		SetPageReserved(vmalloc_to_page((void *)adr));
#else
		mem_map_reserve(vmalloc_to_page((void *)adr));
#endif
		adr+=PAGE_SIZE;
		size-=PAGE_SIZE;
	}
  }
  return mem;
}

static void rvfree(void * mem, unsigned long size) {
  unsigned long adr;

  if (mem) {
	adr=(unsigned long) mem;
	while ((long) size > 0) {
#ifdef KERN26_IFACE
		ClearPageReserved(vmalloc_to_page((void *)adr));
#else
		mem_map_unreserve(vmalloc_to_page((void *)adr));
#endif
		adr+=PAGE_SIZE;
		size-=PAGE_SIZE;
	}
        vfree(mem);
  }
}

#ifdef KERN26_IFACE
int uscam_video_mmap( struct file *file, struct vm_area_struct *vma ) {
  unsigned long start = vma->vm_start;
  unsigned long size = vma->vm_end - vma->vm_start;
#else
#ifdef OLD_IFACE
int uscam_video_mmap( struct video_device *dev, const char *addr, unsigned long size ) {
#else
int uscam_video_mmap( struct vm_area_struct *vma, struct video_device *dev, const char *addr, unsigned long size ) {
#endif
  unsigned long start = (unsigned long)addr;
#endif
  unsigned long page, pos;
  
  pos = (unsigned long) framebuf;
  while (size > 0) {
	page = kvirt_to_pa(pos);
#ifdef OLD_IFACE
	if ( remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED) )
#else
	if ( remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED) )
#endif
		return -EAGAIN;

	start += PAGE_SIZE;
	pos += PAGE_SIZE;
	if ( size > PAGE_SIZE ) size -= PAGE_SIZE; else size = 0;
  }
  
  return 0;
}

#ifdef KERN26_IFACE
static struct file_operations uscam_fops = {
  .owner =	THIS_MODULE,
  .open =	uscam_video_open,
  .release =	uscam_video_close,
  .read =	uscam_video_read,
  .poll =	uscam_video_poll,
  .mmap =	uscam_video_mmap,
  .ioctl =	uscam_video_ioctl,
  .llseek = 	no_llseek,
};

static struct video_device uscam_dev = {
  .owner =	THIS_MODULE,
  .name =	"User-space camera",
  .type =	VID_TYPE_CAPTURE,
  .hardware =	0,
  .fops =	&uscam_fops,
};
#else
static struct video_device uscam_dev = {
  .owner =	THIS_MODULE,
  .name	=	"User-space camera",
  .type =	VID_TYPE_CAPTURE,
  .hardware =	0,
  .open =	uscam_video_open,
  .close =	uscam_video_close,
  .read =	uscam_video_read,
  .poll =	uscam_video_poll,
  .ioctl =	uscam_video_ioctl,
  .mmap =	uscam_video_mmap,
  .initialize =	0,
  .minor =	0
};
#endif

int uscam_proc_read2( struct file *f, char *buf, size_t count, loff_t *pos ) {
  char lbuf[200];
  int written, off = *pos;

  written = sprintf( lbuf, "active: %s\n", activated? "yes": "no" );
  written += sprintf( lbuf + written, "xsize: %d\n", mdev_info.xsize );
  written += sprintf( lbuf + written, "ysize: %d\n", mdev_info.ysize );
  written += sprintf( lbuf + written, "framelen: %d\n", mdev_info.framelen );
  written += sprintf( lbuf + written, "format: %d\n", mdev_info.format );
  if ( written < off ) {
  	written = 0;
	off -= written;
  }
  written -= off;
  if ( written > count ) {
  	copy_to_user( buf + off, lbuf, count );
	*pos += count;
	return count;
  }
  if ( written < 0 ) written = 0;

  copy_to_user( buf + off, lbuf, written );
  *pos += written;

  return written;
}

struct uscam_dev_priv {
  char *buf;
  int bsize;
};

int uscam_proc_mmap( struct file *f, struct vm_area_struct *vmarea ) {
  unsigned long start = vmarea->vm_start;
  unsigned long size = vmarea->vm_end - vmarea->vm_start;
  unsigned long page, pos;
  
  pos = (unsigned long) framebuf;
  while (size > 0) {
	page = kvirt_to_pa(pos);
#ifdef OLD_IFACE
	if ( remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED) )
#else
	if ( remap_page_range(vmarea, start, page, PAGE_SIZE, PAGE_SHARED) )
#endif
		return -EAGAIN;

	start += PAGE_SIZE;
	pos += PAGE_SIZE;
	if ( size > PAGE_SIZE ) size -= PAGE_SIZE; else size = 0;
  }
  
  return 0;
}

#define USCAM_SET_MDEV_INFO	0x1301
#define USCAM_ACTIVATE		0x1302
#define USCAM_FRAME_READY	0x1303

int uscam_proc_ioctl( struct inode *ind, struct file *f,
	unsigned int cmd, unsigned long arg ) {
  struct uscam_mdev_info old_info;
  int activate;

  switch ( cmd ) {
  case USCAM_SET_MDEV_INFO:
	old_info = mdev_info;
	if ( copy_from_user( &mdev_info, (void*) arg, sizeof(mdev_info) ) )
		return -EFAULT;
	if ( old_info.framelen != mdev_info.framelen ) {
	  	rvfree( framebuf, old_info.framelen );
		framebuf = rvmalloc( mdev_info.framelen );
	}
	break;
  case USCAM_ACTIVATE:
	if ( copy_from_user( &activate, (void*) arg, sizeof(activate) ) )
		return -EFAULT;
	if ( activate && !activated ) {
		int i, video_nr = -1;

  		i = video_register_device( &uscam_dev, VFL_TYPE_GRABBER, video_nr );
		if ( i < 0 ) {
			printk(KERN_ERR "Failed to register as video device (%d).\n", i );
			return -EFAULT;
		}
  
		printk(KERN_INFO __FILE__": registered as /dev/video%d.\n",
  			uscam_dev.minor & 0x3F);
	}
	else if ( !activate && activated )
		video_unregister_device( &uscam_dev );
	activated = activate;
  	break;
  case USCAM_FRAME_READY:
  	has_frame = 1;
  	wake_up_interruptible( &frameq );
	break;
  default:
  	return -ENOIOCTLCMD;
  }

  return 0;
}

struct proc_dir_entry *uscam_proc;

struct file_operations uscam_pfops = {
read: uscam_proc_read2,
mmap: uscam_proc_mmap,
ioctl: uscam_proc_ioctl,
};

int __init uscam_init(void) {
//  uscam_dev_priv *priv;

//  priv = (uscam_dev_priv*) vmalloc_32( sizeof(uscam_dev_priv) );
//  memset( priv, 0, sizeof(priv) );
//  video_dev->priv = priv;

  mdev_info.xsize = 320;
  mdev_info.ysize = 240;
  mdev_info.framelen = 320 * 240 * 3;
  mdev_info.format = VIDEO_PALETTE_RGB24;

  framebuf = rvmalloc( mdev_info.framelen );

  uscam_proc = create_proc_entry( "driver/uscam", 0444, 0 );
  uscam_proc->proc_fops = &uscam_pfops;

  has_frame = 0;
  init_waitqueue_head( &frameq );
  activated = 0;

  return 0;
}

void __exit uscam_exit(void) {
  if ( activated ) video_unregister_device( &uscam_dev );
//  vfree( video_dev->priv );
  remove_proc_entry( "driver/uscam", 0 );
  rvfree( framebuf, mdev_info.framelen );
}

MODULE_DESCRIPTION("User-space camera driver");
MODULE_AUTHOR("Vad Rulezz, <vr5@narod.ru>");
MODULE_LICENSE("GPL");

module_init( uscam_init );
module_exit( uscam_exit );
